from typing import List, Union
from dataclasses import dataclass
from enum import Enum


class DifferenceTypes(Enum):
    NONE = 0
    MISMATCH = 1
    S1_MISSING = 2
    S2_MISSING = 3

@dataclass
class Difference:
    index: int
    diff_type: DifferenceTypes

    def __repr__(self) -> str:
        return f'<Difference index={self.index} diff_type={self.diff_type}>'

@dataclass
class Differences:
    sentence_labels_1: List[str]
    sentence_labels_2: List[str]
    diffs_between_labels: List[Difference]

    @property
    def has_diffs(self) -> bool:
        return len(self.diffs_between_labels) > 0

EMPTY_SET = ['O', 'NO-RELATION']

def check_for_differences(l1: Union[List[str], str], l2: Union[List[str], str]) -> Differences:
    def get_difference_type(label_1: str, label_2: str) -> DifferenceTypes:
        diff_type = DifferenceTypes.NONE
        if label_2 in EMPTY_SET or len(label_2) == 0:
            diff_type = DifferenceTypes.S2_MISSING
        elif label_1 in EMPTY_SET or len(label_1) == 0:
            diff_type = DifferenceTypes.S1_MISSING
        else:
            diff_type = DifferenceTypes.MISMATCH

        return diff_type

    labels_1 = [l1] if isinstance(l1, str) else l1
    labels_2 = [l2] if isinstance(l2, str) else l2

    assert len(labels_1) == len(labels_2)

    diffs_between_labels = []
    for i, [label_1, label_2] in enumerate(zip(labels_1, labels_2)):
        if label_1 != label_2:
            diff_type = get_difference_type(label_1, label_2)
            diffs_between_labels.append(Difference(i, diff_type))

    return Differences(
        labels_1,
        labels_2,
        diffs_between_labels
    )



from typing import Callable, Dict, List, Tuple
from extr import Relation, Entity, TokenGroup
from extr.entities import EntityExtractor, \
                          EntityAnnotator
from extr.relations import RelationExtractor, RelationAnnotator
from extr.tokenizers import tokenizer
from extr.utils import Query
from ..models import RelationLabel


class RelationClassification:
    def __init__(self,
                 sentence_tokenizer: Callable[[str], List[List[str]]],
                 entity_extractor: EntityExtractor,
                 relation_extractor: RelationExtractor,
                 no_relations: List[Tuple[str, str, str]],
                 relation_annotator = RelationAnnotator()):
        self._sentence_tokenizer = sentence_tokenizer
        self._entity_extractor = entity_extractor
        self._relation_extractor = relation_extractor
        self._no_relations = no_relations

        self._entities_we_care_about = set()
        for e1, e2, _ in self._no_relations:
            self._entities_we_care_about.add(e1)
            self._entities_we_care_about.add(e2)

        self._entity_annotator = EntityAnnotator()
        self._relation_annotator = relation_annotator

    def label(self, text: str) -> List[RelationLabel]:
        found_entities = Query(self._entity_extractor.get_entities(text)) \
            .filter(lambda entity: entity.label in self._entities_we_care_about) \
            .tolist()

        found_relations = self._relation_extractor.extract(
            self._entity_annotator.annotate(text, found_entities)
        )

        def get_entities_in_token_group(token_group: TokenGroup) -> List[Entity]:
            return Query(found_entities) \
                .filter(token_group.contains) \
                .tolist()

        def get_relations_in_token_group(token_group: TokenGroup) -> List[Relation]:
            return Query(found_relations) \
                .filter(lambda relation: (
                    token_group.contains(relation.e1) and \
                    token_group.contains(relation.e2)
                )) \
                .tolist()

        relation_labels: List[RelationLabel] = []
        for token_group in tokenizer(text, self._sentence_tokenizer(text)):
            relation_mapping = self._get_mappings(
                get_entities_in_token_group(token_group),
                get_relations_in_token_group(token_group)
            )

            offset = token_group.location.start
            for relation in relation_mapping.values():
                relation_labels.append(
                    RelationLabel(
                        self._relation_annotator.annotate(
                            token_group.sentence,
                            relation,
                            offset
                        ).strip(),
                        relation
                    )
                )

        return relation_labels

    def _get_mappings(self, entities_in_sentence: List[Entity], relations_in_sentence: List[Relation]) -> Dict[str, Relation]:
        def get_entities_for_label(label: str) -> List[Entity]:
            return Query(entities_in_sentence) \
                .filter(lambda e: e.label == label) \
                .tolist()

        relation_mapping: Dict[str, Relation] = {}
        for entity1_label, entity2_label, label in self._no_relations:
            e1s = get_entities_for_label(entity1_label)
            e2s = get_entities_for_label(entity2_label)
            for e1 in e1s:
                for e2 in e2s:
                    key = Relation.create_key(e1, e2)
                    relation_mapping[key] = Relation(
                        label,
                        e1,
                        e2
                    )

        for relation in relations_in_sentence:
            relation_mapping[relation.key] = relation

        return relation_mapping



from typing import Dict, List
import re
import os
import json

from extr.relations import HtmlRelationAnnotator
from extr_ds.models import RelationLabel

from ..workspace import WORKSPACE
from ...utils.filesystem import load_document, save_document


HTML_PATH = os.path.join(WORKSPACE, '3', 'dev-rels.html')
DEV_PATH = os.path.join(WORKSPACE, '3', 'dev-rels.json')

def create_html_file(rows: List[str]) -> str:
    def get_default_styles() -> str:
        return """
span.entity { border: 1px solid black; border-radius: 5px; padding: 5px; margin: 3px; cursor: pointer; }
span.label { font-weight: bold; padding: 3px; color: black; }
span.e1 { background-color: aqua; }
span.e2 { background-color: coral; }
tr.delete { background-color: black !important; }
tr.delete span { background-color: black !important; }
td { line-height: 30px; border: 1px solid black; padding: 5px; }
td.header { font-weight: bold; }
td.label { font-weight: bold; text-align: center; }
"""

    return """
<html>
    <head>
        <style>""" + get_default_styles() + """</style>
    </head>
    <body><table>""" + '\n'.join(rows) + """</table></body>
</html>
"""

def create_dev_files(relation_groups: Dict[str, List[RelationLabel]]) -> None:
    html_rows: List[str] = []
    ordered_relations: List[RelationLabel] = []

    index = 0
    html_annotator = HtmlRelationAnnotator()
    for key, items in relation_groups.items():
        html_rows.append(f'<tr><td class="header" colspan=3>{key}</td></tr>')
        for relation_label in items:
            text = html_annotator.annotate(
                relation_label.original_sentence,
                relation_label.relation
            )

            html_rows.append(
                f'<tr id="{index}"><td>{index}</td><td class="label">{relation_label.relation.label}</td><td>{text}</td></tr>'
            )

            ordered_relations.append(relation_label)

            index += 1

    save_document(
        DEV_PATH,
        json.dumps(
            list(map(lambda relation: relation.todict(), ordered_relations)),
            indent=2
        )
    )

    save_document(
        HTML_PATH,
        create_html_file(html_rows)
    )

def change_label(label: str, rows: List[int]) -> None:
    html = load_document(HTML_PATH)
    for row in rows:
        html = re.sub(
            r'(<tr id="' + str(row) + '"><td>' + str(row) + '</td><td class="label">)(.+?)(</td>)',
            r'\1' + label + r'\3',
            html
        )

    save_document(HTML_PATH, html)

    dev = json.loads(load_document(DEV_PATH))
    for row in rows:
        dev[row] = {
            'sentence': dev[row]['sentence'],
            'label': label,
            'definition': dev[row]['definition'],
        }

    save_document(DEV_PATH, json.dumps(dev, indent=2))

def delete_row(rows: List[int]) -> None:
    html = load_document(HTML_PATH)
    for row in rows:
        html = re.sub(
            r'(<tr )(id="' + str(row) + '")',
            r'\1class="delete" \2',
            html
        )

    save_document(HTML_PATH, html)

    dev = json.loads(load_document(DEV_PATH))
    for row in rows:
        dev[row] = {
            'sentence': dev[row]['sentence'],
            'label': dev[row]['label'],
            'definition': dev[row]['definition'],
            'attribute': 'delete',
        }

    save_document(DEV_PATH, json.dumps(dev, indent=2))

def recover_row(rows: List[int]) -> None:
    html = load_document(HTML_PATH)
    for row in rows:
        html = re.sub(
            r'(<tr) class="delete" (id="' + str(row) + '")',
            r'\1 \2',
            html
        )

    save_document(HTML_PATH, html)

    dev = json.loads(load_document(DEV_PATH))
    for row in rows:
        dev[row] = {
            'sentence': dev[row]['sentence'],
            'label': dev[row]['label'],
            'definition': dev[row]['definition'],
        }

    save_document(DEV_PATH, json.dumps(dev, indent=2))



from typing import Any, Dict

import re
import os
import json

from .workspace import WORKSPACE
from ..utils.filesystem import load_data, \
                               append_data, \
                               save_data, \
                               save_document


def entity_text_annotation_to_json(text_annotation: str) -> Dict[str, Any]:
    blob = {
        'text': '(<TIME>0:26</TIME> - <PERIOD>2nd</PERIOD>) Timeout #3 by <TEAM>MIA</TEAM> at <TIME>00:26</TIME>.',
        'entities': []
    }

    while True:

        match = re.search(r'(<(\w+?)>.+?</\w+>)', blob['text'])
        if match is None:
            break

        start, match_end = match.span()
        entity_text = re.sub(r'</?\w+>', '', match.group(1))
        label = match.group(2)


        blob['entities'].append(
            {
                'start': start,
                'end': start + len(entity_text),
                'label': label,
                'text': entity_text
            }
        )

        blob['text'] = blob['text'][:start] + entity_text + blob['text'][match_end:]

    return blob


def save_entities() -> None:
    blobs = [
        entity_text_annotation_to_json(text)
        for text in load_data(os.path.join(WORKSPACE, '3', 'dev-ents.txt'))
    ]

    current_size = 0

    blob_storage = os.path.join(WORKSPACE, '4', 'ents.json')
    if os.path.exists(blob_storage):
        with open(blob_storage, 'r', encoding='utf-8') as blob_outputs:
            current_data = json.loads(blob_outputs.read())

        current_size = len(current_data)
        blobs.extend(current_data)

    keys = set()
    dataset = []
    for blob in blobs:
        key = blob['sentence']
        if key in keys:
            continue

        keys.add(key)
        dataset.append(blob)

    save_data(
        os.path.join(WORKSPACE, '4', 'ents.txt'),
        list(map(lambda item: item['sentence'], dataset))
    )

    print('#rows:', len(dataset) - current_size)

    save_document(
        blob_storage,
        json.dumps(dataset, indent=2)
    )

    redacted_dataset = set(load_data(os.path.join(WORKSPACE, '3', 'dev-ents-redacted.txt')))
    append_data(os.path.join(WORKSPACE, '4', 'ents-redacted.txt'), redacted_dataset)

def save_relations() -> None:
    with open(os.path.join(WORKSPACE, '3', 'dev-rels.json'), 'r', encoding='utf-8') as relation_outputs:
        data = [row for row in json.loads(relation_outputs.read()) if not 'attribute' in row]

    output_path = os.path.join(WORKSPACE, '4', 'rels.json')
    if os.path.exists(output_path):
        with open(output_path, 'r', encoding='utf-8') as relation_outputs:
            current_data = json.loads(relation_outputs.read())

        data.extend(current_data)

    lookup: Dict[str, Dict[str, Any]] = {}
    for row in data:
        key = row['sentence'].lower()
        if not key in lookup:
            lookup[key] = { 'labels': set(), 'items': [] }

        lookup[key]['labels'].add(row['label'])
        lookup[key]['items'].append(row)

    slim_data = []
    for key, item in lookup.items():
        labels = item['labels']
        items = item['items']

        if len(labels) == 1:
            slim_data.append(items[0])
        else:
            print('Label mismatch found (skipping):')
            print('sentence:', items[0]['sentence'])
            print('labels:', ', '.join(labels))
            print()

    save_document(
        output_path,
        json.dumps(slim_data, indent=2)
    )

    print('#rows:', len(slim_data))
